@@ -609,8 +609,7 @@ bool pkgPackageManager::SmartConfigure(PkgIterator Pkg, int const Depth)
 
    List->Flag(Pkg,pkgOrderList::Configured,pkgOrderList::States);
 
-   if ((Cache[Pkg].InstVerIter(Cache)->MultiArch & pkgCache::Version::Same) == pkgCache::Version::Same &&
-       not List->IsFlag(Pkg, pkgOrderList::Immediate))
+   if ((Cache[Pkg].InstVerIter(Cache)->MultiArch & pkgCache::Version::Same) == pkgCache::Version::Same)
       for (PkgIterator P = Pkg.Group().PackageList();
 	   P.end() == false; P = Pkg.Group().NextPkg(P))
       {
@@ -652,11 +651,9 @@ bool pkgPackageManager::EarlyRemove(PkgIterator Pkg, DepIterator const * const D
 
    // Essential packages get special treatment
    bool IsEssential = false;
-   if ((Pkg->Flags & pkgCache::Flag::Essential) != 0)
+   if ((Pkg->Flags & pkgCache::Flag::Essential) != 0 ||
+       (Pkg->Flags & pkgCache::Flag::Important) != 0)
       IsEssential = true;
-   bool IsProtected = false;
-   if ((Pkg->Flags & pkgCache::Flag::Important) != 0)
-      IsProtected = true;
 
    /* Check for packages that are the dependents of essential packages and
       promote them too */
@@ -664,17 +661,14 @@ bool pkgPackageManager::EarlyRemove(PkgIterator Pkg, DepIterator const * const D
    {
       for (pkgCache::DepIterator D = Pkg.RevDependsList(); D.end() == false &&
 	   IsEssential == false; ++D)
-	 if (D->Type == pkgCache::Dep::Depends || D->Type == pkgCache::Dep::PreDepends) {
-	    if ((D.ParentPkg()->Flags & pkgCache::Flag::Essential) != 0)
+	 if (D->Type == pkgCache::Dep::Depends || D->Type == pkgCache::Dep::PreDepends)
+	    if ((D.ParentPkg()->Flags & pkgCache::Flag::Essential) != 0 ||
+	        (D.ParentPkg()->Flags & pkgCache::Flag::Important) != 0)
 	       IsEssential = true;
-	    if ((D.ParentPkg()->Flags & pkgCache::Flag::Important) != 0)
-	       IsProtected = true;
-	 }
    }
 
    if (IsEssential == true)
    {
-      // FIXME: Unify messaging with Protected below.
       if (_config->FindB("APT::Force-LoopBreak",false) == false)
 	 return _error->Error(_("This installation run will require temporarily "
 				"removing the essential package %s due to a "
@@ -685,16 +679,6 @@ bool pkgPackageManager::EarlyRemove(PkgIterator Pkg, DepIterator const * const D
    // dpkg will auto-deconfigure it, no need for the big remove hammer
    else if (Dep != NULL && (*Dep)->Type == pkgCache::Dep::DpkgBreaks)
       return true;
-   else if (IsProtected == true)
-   {
-      // FIXME: Message should talk about Protected, not Essential, and unified.
-      if (_config->FindB("APT::Force-LoopBreak",false) == false)
-	 return _error->Error(_("This installation run will require temporarily "
-				"removing the essential package %s due to a "
-				"Conflicts/Pre-Depends loop. This is often bad, "
-				"but if you really want to do it, activate the "
-				"APT::Force-LoopBreak option."),Pkg.FullName().c_str());
-   }
 
    bool Res = SmartRemove(Pkg);
    if (Cache[Pkg].Delete() == false)
@@ -1021,18 +1005,10 @@ bool pkgPackageManager::SmartUnPack(PkgIterator Pkg, bool const Immediate, int c
       return false;
 
    if (Immediate == true) {
-      // Perform immediate configuration of the package.
-      _error->PushToStack();
-      bool configured = SmartConfigure(Pkg, Depth + 1);
-      _error->RevertToStack();
-
-      if (not configured && Debug) {
-	 clog << OutputInDepth(Depth);
-	 ioprintf(clog, _("Could not perform immediate configuration on '%s'. "
-			   "Please see man 5 apt.conf under APT::Immediate-Configure for details. (%d)"),
-			 Pkg.FullName().c_str(), 2);
-	 clog << endl;
-      }
+      // Perform immediate configuration of the package. 
+         if (SmartConfigure(Pkg, Depth + 1) == false)
+            _error->Warning(_("Could not perform immediate configuration on '%s'. "
+               "Please see man 5 apt.conf under APT::Immediate-Configure for details. (%d)"),Pkg.FullName().c_str(),2);
    }
    
    return true;
